from __future__ import annotations


class Event:
    """
    Event
    """

    def __init__(self) -> None:
        self._propagation_stopped = False

    def is_propagation_stopped(self) -> bool:
        return self._propagation_stopped

    def stop_propagation(self) -> None:
        self._propagation_stopped = True



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Union


if TYPE_CHECKING:
    import sys

    if sys.version_info >= (3, 8):
        from typing import Literal
    else:
        from typing_extensions import Literal

    _Align = Union[Literal["left"], Literal["right"]]


class TableCellStyle:
    def __init__(
        self,
        fg: str = "default",
        bg: str = "default",
        options: list[str] | None = None,
        align: _Align = "left",
        cell_format: str | None = None,
    ) -> None:
        self._fg = fg
        self._bg = bg
        self._options = options
        self._align = "left"
        self._cell_format = cell_format

    @property
    def cell_format(self) -> str | None:
        return self._cell_format

    @property
    def tag(self) -> str:
        tag = "<fg={};bg={}"

        if self._options:
            tag += f';options={",".join(self._options)}'

        tag += ">"

        return tag

    def pad(self, string: str, length: int, char: str = " ") -> str:
        if self._align == "left":
            return string.rjust(length, char)

        if self._align == "right":
            return string.ljust(length, char)

        return string.center(length, char)



from __future__ import annotations

from io import StringIO
from typing import TYPE_CHECKING

from cleo.io.buffered_io import BufferedIO
from cleo.io.inputs.string_input import StringInput
from cleo.io.outputs.buffered_output import BufferedOutput


if TYPE_CHECKING:
    from cleo.application import Application
    from cleo.io.outputs.output import Verbosity


class ApplicationTester:
    """
    Eases the testing of console applications.
    """

    def __init__(self, application: Application) -> None:
        self._application = application
        self._application.auto_exits(False)
        self._io = BufferedIO()
        self._status_code = 0

    @property
    def application(self) -> Application:
        return self._application

    @property
    def io(self) -> BufferedIO:
        return self._io

    @property
    def status_code(self) -> int:
        return self._status_code

    def execute(
        self,
        args: str = "",
        inputs: str | None = None,
        interactive: bool = True,
        verbosity: Verbosity | None = None,
        decorated: bool = False,
        supports_utf8: bool = True,
    ) -> int:
        """
        Executes the command
        """
        self._io.clear()

        input = StringInput(args)
        self._io.set_input(input)
        self._io.decorated(decorated)
        assert isinstance(self._io.output, BufferedOutput)
        assert isinstance(self._io.error_output, BufferedOutput)
        self._io.output.set_supports_utf8(supports_utf8)
        self._io.error_output.set_supports_utf8(supports_utf8)

        if inputs is not None:
            self._io.input.set_stream(StringIO(inputs))

        if interactive is not None:
            self._io.interactive(interactive)

        if verbosity is not None:
            self._io.set_verbosity(verbosity)

        self._status_code = self._application.run(
            self._io.input,
            self._io.output,
            self._io.error_output,
        )

        return self._status_code



from __future__ import annotations

from cleo.color import Color


class Style:
    def __init__(
        self,
        foreground: str | None = None,
        background: str | None = None,
        options: list[str] | None = None,
    ) -> None:
        self._foreground = foreground or ""
        self._background = background or ""
        self._options = options or []

        self._color = Color(self._foreground, self._background, self._options)

    def foreground(self, foreground: str) -> Style:
        self._color = Color(foreground, self._background, self._options)
        self._foreground = foreground

        return self

    def background(self, background: str) -> Style:
        self._color = Color(self._foreground, background, self._options)
        self._background = background

        return self

    def bold(self, bold: bool = True) -> Style:
        return self.set_option("bold") if bold else self.unset_option("bold")

    def dark(self, dark: bool = True) -> Style:
        return self.set_option("dark") if dark else self.unset_option("dark")

    def underlines(self, underlined: bool = True) -> Style:
        return (
            self.set_option("underline")
            if underlined
            else self.unset_option("underline")
        )

    def italic(self, italic: bool = True) -> Style:
        return self.set_option("italic") if italic else self.unset_option("italic")

    def blinking(self, blinking: bool = True) -> Style:
        return self.set_option("blink") if blinking else self.unset_option("blink")

    def inverse(self, inverse: bool = True) -> Style:
        return self.set_option("reverse") if inverse else self.unset_option("reverse")

    def hidden(self, hidden: bool = True) -> Style:
        return self.set_option("conceal") if hidden else self.unset_option("conceal")

    def set_option(self, option: str) -> Style:
        self._options.append(option)
        self._color = Color(self._foreground, self._background, self._options)

        return self

    def unset_option(self, option: str) -> Style:
        if option in self._options:
            index = self._options.index(option)
            del self._options[index]
            self._color = Color(self._foreground, self._background, self._options)
        return self

    def apply(self, text: str) -> str:
        return self._color.apply(text)



from __future__ import annotations


class TokenParser:
    """
    Parses tokens from a string passed to StringArgs.
    """

    def __init__(self) -> None:
        self._string: str = ""
        self._cursor: int = 0
        self._current: str | None = None
        self._next_: str | None = None

    def parse(self, string: str) -> list[str]:
        self._string = string
        self._cursor = 0
        self._current = None
        if len(string) > 0:
            self._current = string[0]

        self._next_ = None
        if len(string) > 1:
            self._next_ = string[1]

        tokens = self._parse()

        return tokens

    def _parse(self) -> list[str]:
        tokens = []

        while self._current is not None:
            if self._current.isspace():
                # Skip spaces
                self._next()

                continue

            if self._current is not None:
                tokens.append(self._parse_token())

        return tokens

    def _next(self) -> None:
        """
        Advances the cursor to the next position.
        """
        if self._current is None:
            return

        self._cursor += 1
        self._current = self._next_

        if self._cursor + 1 < len(self._string):
            self._next_ = self._string[self._cursor + 1]
        else:
            self._next_ = None

    def _parse_token(self) -> str:
        token = ""

        while self._current is not None:
            if self._current.isspace():
                self._next()

                break

            if self._current == "\\":
                token += self._parse_escape_sequence()
            elif self._current in ["'", '"']:
                token += self._parse_quoted_string()
            else:
                token += self._current
                self._next()

        return token

    def _parse_quoted_string(self) -> str:
        string = ""
        delimiter = self._current

        # Skip first delimiter
        self._next()
        while self._current is not None:
            if self._current == delimiter:
                # Skip last delimiter
                self._next()

                break

            if self._current == "\\":
                string += self._parse_escape_sequence()
            elif self._current == '"':
                string += f'"{self._parse_quoted_string()}"'
            elif self._current == "'":
                string += f"'{self._parse_quoted_string()}'"
            else:
                string += self._current
                self._next()

        return string

    def _parse_escape_sequence(self) -> str:
        if self._next_ in ['"', "'"]:
            sequence = self._next_
        else:
            assert self._next_ is not None
            sequence = "\\" + self._next_

        self._next()
        self._next()

        return sequence



from __future__ import annotations

import os

from cleo.exceptions import CleoValueError


class Color:

    COLORS = {
        "black": (30, 40),
        "red": (31, 41),
        "green": (32, 42),
        "yellow": (33, 43),
        "blue": (34, 44),
        "magenta": (35, 45),
        "cyan": (36, 46),
        "light_gray": (37, 47),
        "default": (39, 49),
        "dark_gray": (90, 100),
        "light_red": (91, 101),
        "light_green": (92, 102),
        "light_yellow": (93, 103),
        "light_blue": (94, 104),
        "light_magenta": (95, 105),
        "light_cyan": (96, 106),
        "white": (97, 107),
    }

    AVAILABLE_OPTIONS = {
        "bold": {"set": 1, "unset": 22},
        "dark": {"set": 2, "unset": 22},
        "italic": {"set": 3, "unset": 23},
        "underline": {"set": 4, "unset": 24},
        "blink": {"set": 5, "unset": 25},
        "reverse": {"set": 7, "unset": 27},
        "conceal": {"set": 8, "unset": 28},
    }

    def __init__(
        self,
        foreground: str = "",
        background: str = "",
        options: list[str] | None = None,
    ) -> None:
        self._foreground = self._parse_color(foreground, False)
        self._background = self._parse_color(background, True)

        if options is None:
            options = []

        self._options = {}
        for option in options:
            if option not in self.AVAILABLE_OPTIONS:
                raise ValueError(
                    f'"{option}" is not a valid color option. '
                    f'It must be one of {", ".join(self.AVAILABLE_OPTIONS.keys())}'
                )

            self._options[option] = self.AVAILABLE_OPTIONS[option]

    def apply(self, text: str) -> str:
        return self.set() + text + self.unset()

    def set(self) -> str:
        codes = []

        if self._foreground:
            codes.append(self._foreground)

        if self._background:
            codes.append(self._background)

        for option in self._options.values():
            codes.append(str(option["set"]))

        if not codes:
            return ""

        return f'\033[{";".join(codes)}m'

    def unset(self) -> str:
        codes = []

        if self._foreground:
            codes.append("39")

        if self._background:
            codes.append("49")

        for option in self._options.values():
            codes.append(str(option["unset"]))

        if not codes:
            return ""

        return f'\033[{";".join(codes)}m'

    def _parse_color(self, color: str, background: bool) -> str:
        if not color:
            return ""

        if color.startswith("#"):
            color = color[1:]

            if len(color) == 3:
                color = color[0] * 2 + color[1] * 2 + color[2] * 2

            if len(color) != 6:
                raise CleoValueError(f'"{color}" is an invalid color')

            return ("4" if background else "3") + self._convert_hex_color_to_ansi(
                int(color, 16)
            )

        if color not in self.COLORS:
            raise CleoValueError(
                f'"{color}" is an invalid color.'
                f' It must be one of {", ".join(self.COLORS.keys())}'
            )

        return str(self.COLORS[color][int(background)])

    def _convert_hex_color_to_ansi(self, color: int) -> str:
        r = (color >> 16) & 255
        g = (color >> 8) & 255
        b = color & 255

        if os.getenv("COLORTERM") != "truecolor":
            return str(self._degrade_hex_color_to_ansi(r, g, b))

        return f"8;2;{r};{g};{b}"

    def _degrade_hex_color_to_ansi(self, r: int, g: int, b: int) -> int:
        if round(self._get_saturation(r, g, b) / 50) == 0:
            return 0

        return (round(b / 255) << 2) | (round(g / 255) << 1) | round(r / 255)

    def _get_saturation(self, r: int, g: int, b: int) -> int:
        r_float = r / 255
        g_float = g / 255
        b_float = b / 255
        v = max(r_float, g_float, b_float)

        diff = v - min(r_float, g_float, b_float)
        if diff == 0:
            return 0

        return int(diff * 100 / v)



from __future__ import annotations


class TableStyle:
    """
    Defines styles for Table instances.
    """

    def __init__(self) -> None:
        self._padding_char = " "
        self._horizontal_outside_border_char = "-"
        self._horizontal_inside_border_char = "-"
        self._vertical_outside_border_char = "|"
        self._vertical_inside_border_char = "|"
        self._crossing_char = "+"
        self._crossing_top_right_char = "+"
        self._crossing_top_mid_char = "+"
        self._crossing_top_left_char = "+"
        self._crossing_mid_right_char = "+"
        self._crossing_bottom_right_char = "+"
        self._crossing_bottom_mid_char = "+"
        self._crossing_bottom_left_char = "+"
        self._crossing_mid_left_char = "+"
        self._crossing_top_left_bottom_char = "+"
        self._crossing_top_mid_bottom_char = "+"
        self._crossing_top_right_bottom_char = "+"
        self._header_title_format = "<b> {} </b>"
        self._footer_title_format = "<b> {} </b>"
        self._cell_header_format = "<c1>{}</c1>"
        self._cell_row_format = "{}"
        self._cell_row_content_format = " {} "
        self._border_format = "{}"
        self._pad_type = "right"

    @property
    def padding_char(self) -> str:
        return self._padding_char

    @property
    def border_chars(self) -> list[str]:
        return [
            self._horizontal_outside_border_char,
            self._vertical_outside_border_char,
            self._horizontal_inside_border_char,
            self._vertical_inside_border_char,
        ]

    @property
    def crossing_char(self) -> str:
        return self._crossing_char

    @property
    def crossing_chars(self) -> list[str]:
        return [
            self._crossing_char,
            self._crossing_top_left_char,
            self._crossing_top_mid_char,
            self._crossing_top_right_char,
            self._crossing_mid_right_char,
            self._crossing_bottom_right_char,
            self._crossing_bottom_mid_char,
            self._crossing_bottom_left_char,
            self._crossing_mid_left_char,
            self._crossing_top_left_bottom_char,
            self._crossing_top_mid_bottom_char,
            self._crossing_top_right_bottom_char,
        ]

    @property
    def cell_header_format(self) -> str:
        return self._cell_header_format

    @property
    def cell_row_format(self) -> str:
        return self._cell_row_format

    @property
    def cell_row_content_format(self) -> str:
        return self._cell_row_content_format

    @property
    def border_format(self) -> str:
        return self._border_format

    @property
    def header_title_format(self) -> str:
        return self._header_title_format

    @property
    def footer_title_format(self) -> str:
        return self._footer_title_format

    @property
    def pad_type(self) -> str:
        return self._pad_type

    def set_padding_char(self, padding_char: str) -> TableStyle:
        """
        Sets padding character, used for cell padding.
        """
        if not padding_char:
            raise ValueError("The padding char must not be empty.")

        self._padding_char = padding_char

        return self

    def set_horizontal_border_chars(
        self, outside: str, inside: str | None = None
    ) -> TableStyle:
        """
        Sets horizontal border characters.

        ╔═══════════════╤══════════════════════════╤══════════════════╗
        1 ISBN          2 Title                    │ Author           ║
        ╠═══════════════╪══════════════════════════╪══════════════════╣
        ║ 99921-58-10-7 │ Divine Comedy            │ Dante Alighieri  ║
        ║ 9971-5-0210-0 │ A Tale of Two Cities     │ Charles Dickens  ║
        ║ 960-425-059-0 │ The Lord of the Rings    │ J. R. R. Tolkien ║
        ║ 80-902734-1-6 │ And Then There Were None │ Agatha Christie  ║
        ╚═══════════════╧══════════════════════════╧══════════════════╝
        """
        self._horizontal_outside_border_char = outside
        self._horizontal_inside_border_char = inside if inside is not None else outside

        return self

    def set_vertical_border_chars(
        self, outside: str, inside: str | None = None
    ) -> TableStyle:
        """
        Sets vertical border characters.

        ╔═══════════════╤══════════════════════════╤══════════════════╗
        ║ ISBN          │ Title                    │ Author           ║
        ╠═══════1═══════╪══════════════════════════╪══════════════════╣
        ║ 99921-58-10-7 │ Divine Comedy            │ Dante Alighieri  ║
        ║ 9971-5-0210-0 │ A Tale of Two Cities     │ Charles Dickens  ║
        ╟───────2───────┼──────────────────────────┼──────────────────╢
        ║ 960-425-059-0 │ The Lord of the Rings    │ J. R. R. Tolkien ║
        ║ 80-902734-1-6 │ And Then There Were None │ Agatha Christie  ║
        ╚═══════════════╧══════════════════════════╧══════════════════╝
        """
        self._vertical_outside_border_char = outside
        self._vertical_inside_border_char = inside if inside is not None else outside

        return self

    def set_crossing_chars(
        self,
        cross: str,
        top_left: str,
        top_mid: str,
        top_right: str,
        mid_right: str,
        bottom_right: str,
        bottom_mid: str,
        bottom_left: str,
        mid_left: str,
        top_left_bottom: str | None = None,
        top_mid_bottom: str | None = None,
        top_right_bottom: str | None = None,
    ) -> TableStyle:
        """
        Sets crossing characters.

        Example:

        1═══════════════2══════════════════════════2══════════════════3
        ║ ISBN          │ Title                    │ Author           ║
        8'══════════════0'═════════════════════════0'═════════════════4'
        ║ 99921-58-10-7 │ Divine Comedy            │ Dante Alighieri  ║
        ║ 9971-5-0210-0 │ A Tale of Two Cities     │ Charles Dickens  ║
        8───────────────0──────────────────────────0──────────────────4
        ║ 960-425-059-0 │ The Lord of the Rings    │ J. R. R. Tolkien ║
        ║ 80-902734-1-6 │ And Then There Were None │ Agatha Christie  ║
        7═══════════════6══════════════════════════6══════════════════5
        """
        self._crossing_char = cross
        self._crossing_top_left_char = top_left
        self._crossing_top_mid_char = top_mid
        self._crossing_top_right_char = top_right
        self._crossing_mid_right_char = mid_right
        self._crossing_bottom_right_char = bottom_right
        self._crossing_bottom_mid_char = bottom_mid
        self._crossing_bottom_left_char = bottom_left
        self._crossing_mid_left_char = mid_left
        self._crossing_top_left_bottom_char = (
            top_left_bottom if top_left_bottom is not None else mid_left
        )
        self._crossing_top_mid_bottom_char = (
            top_mid_bottom if top_mid_bottom is not None else cross
        )
        self._crossing_top_right_bottom_char = (
            top_right_bottom if top_right_bottom is not None else mid_right
        )

        return self

    def set_default_crossing_char(self, char: str) -> TableStyle:
        """
        Sets default crossing character used for each cross.
        """
        return self.set_crossing_chars(
            char, char, char, char, char, char, char, char, char
        )

    def set_cell_header_format(self, cell_header_format: str) -> TableStyle:
        """
        Sets the header cell format.
        """
        self._cell_header_format = cell_header_format

        return self

    def set_cell_row_format(self, cell_row_format: str) -> TableStyle:
        """
        Sets the row cell format.
        """
        self._cell_row_format = cell_row_format

        return self

    def set_cell_row_content_format(self, cell_row_content_format: str) -> TableStyle:
        """
        Sets the row cell content format.
        """
        self._cell_row_content_format = cell_row_content_format

        return self

    def set_border_format(self, border_format: str) -> TableStyle:
        """
        Sets the border format.
        """
        self._border_format = border_format

        return self

    def set_header_title_format(self, header_title_format: str) -> TableStyle:
        """
        Sets the header title format.
        """
        self._header_title_format = header_title_format

        return self

    def set_footer_title_format(self, footer_title_format: str) -> TableStyle:
        """
        Sets the footer title format.
        """
        self._footer_title_format = footer_title_format

        return self

    def set_pad_type(self, pad_type: str) -> TableStyle:
        """
        Sets the padding type.
        """
        if pad_type not in {"left", "right", "center"}:
            raise ValueError(
                'Invalid padding type. Expected one of "left", "right", "center").'
            )

        self._pad_type = pad_type

        return self

    def pad(self, string: str, length: int, char: str = " ") -> str:
        if self._pad_type == "left":
            return string.rjust(length, char)

        if self._pad_type == "right":
            return string.ljust(length, char)

        return string.center(length, char)
